@@ -17,6 +17,8 @@ import pandas as pd
 from pyntcloud import PyntCloud
 import ot
 from multipoles import MultipoleExpansion
+import pyvista as pv
+from pcdiff import knn_graph, estimate_basis, build_grad_div, laplacian
 
 def get_border_dict(labels,states=None,radius=10,state_ids=None,vdist=None,return_nnindex=True,return_nnvector=True,return_curvature=True,nn_states=None,scale=None,verbose=False,**border_args):
     """
@@ -469,19 +471,37 @@ def get_surface_displacement(border_dict,sts=None,c=None,n=None,alpha=1.,maxd=No
         dx=dx*(maxd/maxr)
     return dx
 
-def get_surface_gradvariance(border_pts,dx_ot,knn=12,use_eigs=False):
+def get_surface_gradvariance(border_pts,dx_ot,knn=12,use_eigs=False,use_nonspatial=False):
     rdx_ot = np.linalg.norm(dx_ot,axis=1)
-    cloud=PyntCloud(pd.DataFrame(data=border_pts,columns=['x','y','z']))
-    k_neighbors = cloud.get_neighbors(k=knn)
     if use_eigs:
+        cloud=PyntCloud(pd.DataFrame(data=border_pts,columns=['x','y','z']))
+        k_neighbors = cloud.get_neighbors(k=knn)
         cloud.points['x']=dx_ot[:,0]
         cloud.points['y']=dx_ot[:,1]
         cloud.points['z']=dx_ot[:,2]
         ev = cloud.add_scalar_field("eigen_decomposition", k_neighbors=k_neighbors)
         w = np.array([cloud.points[ev[0]],cloud.points[ev[1]],cloud.points[ev[2]]]).T
         dh = np.sum(w,axis=1)
-    else:
+    if use_nonspatial:
+        cloud=PyntCloud(pd.DataFrame(data=border_pts,columns=['x','y','z']))
+        k_neighbors = cloud.get_neighbors(k=knn)
         dh=np.var(rdx_ot[k_neighbors],axis=1)
+    else:
+        rdx_ot = np.linalg.norm(dx_ot,axis=1)
+        dh = np.ones_like(rdx_ot)*np.nan
+        indgood=np.where(np.isfinite(rdx_ot))[0]
+        if indgood.size>0:
+            pts=border_pts[indgood,:]
+            f=rdx_ot[indgood]
+            edge_index = knn_graph(pts, knn)
+            # Estimate normals and local frames
+            basis = estimate_basis(pts, edge_index)
+            # Build gradient and divergence operators (Scipy sparse matrices)
+            grad, div = build_grad_div(border_pts, *basis, edge_index)
+            grad_h = grad @ rdx_ot
+            grad_h = grad_h.reshape(border_pts.shape[0],2)
+            dh = np.ones_like(rdx_ot)*np.nan
+            dh[indgood] = (grad_h * grad_h).sum(axis=1)
     return dh
 
 def get_surface_displacement_deviation(border_dict,border_pts_prev,exclude_states=None,n=None,knn=12,use_eigs=False,alpha=1.,maxd=None):
@@ -1409,7 +1429,7 @@ def get_flux_ligdist(vdist,cmean=1.,csigma=.5,center=True):
     fsigmas=np.abs(csigma*np.abs(vdist))
     return fmeans,fsigmas
 
-def get_border_properties(cell_labels,surfaces=None,cell_states=None,state_ids=None,surface_states=None,cell_labels_next=None,cell_labels_prev=None,lin_next=None,lin_prev=None,vdist=None,border_scale=1.0,radius=2.0,order=1,return_border_properties_list=False):
+def get_border_properties(cell_labels,surfaces=None,cell_states=None,state_ids=None,surface_states=None,cell_labels_next=None,cell_labels_prev=None,lin_next=None,lin_prev=None,vdist=None,border_scale=1.0,radius=2.0,order=1,return_border_properties_list=False,verbose=False):
     """
     Calculate geometric and physical properties of cell borders, with optional tracking between frames 
     and surface contact details.
@@ -1564,7 +1584,8 @@ def get_border_properties(cell_labels,surfaces=None,cell_states=None,state_ids=N
                     border_charge=np.ones(indcell.size)*-1
                     border_charge[indcell_surf]=1
                     av_surf_fraction=indcell_surf.size/indcell.size
-                    print(f'\rcell {cellid} state {i_state} contact fraction {av_surf_fraction:.2f}              ',end="")
+                    if verbose:
+                        print(f'\rcell {cellid} state {i_state} contact fraction {av_surf_fraction:.2f}              ',end="")
                     charges_list = [{'q': q, 'xyz': tuple(border_pts)} for q, border_pts in zip(border_charge, border_pts)]
                     charge_dist = {'discrete': True, 'charges': charges_list}
                     Phi = MultipoleExpansion(charge_dist, order)
@@ -1575,7 +1596,8 @@ def get_border_properties(cell_labels,surfaces=None,cell_states=None,state_ids=N
                             contact_properties[cellid,i_state,i_prop]=Phi.multipole_moments[l,m]
                             i_prop=i_prop+1
                 else:
-                    print(f'\rcell {cellid} state {i_state} contact fraction {0.0:.2f}              ',end="")
+                    if verbose:
+                        print(f'\rcell {cellid} state {i_state} contact fraction {0.0:.2f}              ',end="")
     border_dict['contact_properties']=contact_properties
     border_dict['border_properties_list']=border_properties_list
     if cell_labels_next is not None:
